package be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec_fx.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicDistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicExecuteAnsiblePlaybook;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicOpenflowDataPath;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecWriter;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.text.ParseException;
import java.util.*;

public class FXModelRspec implements ModelRspec {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(FXModelRspec.class);

    final List<ExtraXml> extraXml = new ArrayList<>();
    private final ObservableList<ExperimentBarrierSegment> barrierSegments = FXCollections.observableArrayList();
    private final Map<String, String> origNameSpaceInfo = new HashMap<>();

    private final ObservableList<FXRspecNode> nodes = FXCollections.observableArrayList();
    private final ObservableList<FXRspecLink> links = FXCollections.observableArrayList();
    private final ObservableList<BasicOpenflowDataPath> openflowDataPaths = FXCollections.observableArrayList();
    private final ObservableList<OpenflowSliver> openflowSlivers = FXCollections.observableArrayList();
    //store BasicOpenflowDataPath as we do nothing special with them
    private final ObservableList<Lease> leases = FXCollections.observableArrayList();
    private final ObservableList<FXRspecChannel> channels = FXCollections.observableArrayList();
    private final ObservableList<FXGeantTestbedType> geantTestbedTypes = FXCollections.observableArrayList();
    private final ObservableList<FXAddressPool> addressPools = FXCollections.observableArrayList();
    private final ObservableList<BasicExecuteAnsiblePlaybook> ansibleCookbooks = FXCollections.observableArrayList();
    private final ObservableList<BasicDistributeSshKeypair> distributeSshKeypairs = FXCollections.observableArrayList();
    private final ObservableList<DiskImageList> diskImageLists = FXCollections.observableArrayList();
    private final ObservableList<FlavorList> flavorLists = FXCollections.observableArrayList();

    private final StringProperty expiresDate = new SimpleStringProperty();
    private final ObjectProperty<HardwareTypeInfo> hardwareTypeInfo = new SimpleObjectProperty<HardwareTypeInfo>();


    private final BooleanProperty routableAddressesAvailableUsed = new SimpleBooleanProperty(false);
    private final BooleanProperty routableAddressesConfiguredUsed = new SimpleBooleanProperty(false);
    private final IntegerProperty routableAddressesAvailable = new SimpleIntegerProperty();
    private final IntegerProperty routableAddressesConfigured = new SimpleIntegerProperty();

    private final ObjectProperty<Integer> multiplexFactor = new SimpleObjectProperty<>();

    private String origDefaultNamespace = null;
    private String origSchemaLocation = null;
    private final String rspecType;

    private final ObservableList<UserSpec> sshKeys = FXCollections.observableArrayList();

    private final RspecFactory factory;

    /**
     * Create an empty Rspec
     */
    public FXModelRspec(String rspecType) {
        this.rspecType = rspecType;
        this.factory = RspecFactoryFactory.getRspecFactoryInstance(getModelRspecType());
    }

    @Nonnull
    @Override
    public ModelRspecType getModelRspecType() {
        return ModelRspecType.FX;
    }

    @Override
    public void setParseData(@Nullable String schemaLocation, @Nullable String defaultNamespace, @Nonnull Map<String, String> namespaces) {
        this.origSchemaLocation = schemaLocation;
        this.origDefaultNamespace = defaultNamespace;
        this.origNameSpaceInfo.putAll(namespaces);
    }

    @Override
    public String getSchemaLocation() {
        return origSchemaLocation;
    }

    @Override
    public String getOrigDefaultNamespace() {
        return origDefaultNamespace;
    }

    @Override
    @Nonnull
    public Map<String, String> getNameSpaceInfo() {
        return origNameSpaceInfo;
    }

    @Override
    public void addNode(@Nonnull RspecNode node) {
        Objects.requireNonNull(node, "may not add null node");
        if (!(node instanceof FXRspecNode))
            throw new IllegalArgumentException("may only add FXRspecNode, not " + node.getClass().getName());

        nodes.add((FXRspecNode) node);
    }

    @Override
    public void addLink(@Nonnull RspecLink link) {
        Objects.requireNonNull(link, "may not add null link");
        if (!(link instanceof FXRspecLink))
            throw new IllegalArgumentException("may only add FXRspecLink, not " + link.getClass().getName());

        links.add((FXRspecLink) link);
    }

    @Override
    @Nonnull
    public ObservableList<FXRspecNode> getNodes() {
        return nodes;
    }

    public ObservableList<ExperimentBarrierSegment> getBarrierSegments() {
        return barrierSegments;
    }

    public void addBarrierSegment(ExperimentBarrierSegment b) {
        for (ExperimentBarrierSegment bs : barrierSegments)
            if (bs.getOrderNumber() == b.getOrderNumber())
                throw new RuntimeException("Duplicate order " + b.getOrderNumber());
        barrierSegments.add(b);
    }

    @Override
    @Nonnull
    public ObservableList<FXRspecLink> getLinks() {
        return links;
    }


    @Override
    public void addOpenflowDataPath(@Nonnull OpenflowDataPath dp) {
        //stores BasicOpenflowDataPath as we do nothing special with them
        assert dp instanceof BasicOpenflowDataPath;
        openflowDataPaths.add((BasicOpenflowDataPath) dp);
    }

    @Override
    @Nonnull
    public List<BasicOpenflowDataPath> getOpenflowDataPaths() {
        //stores BasicOpenflowDataPath as we do nothing special with them
        return openflowDataPaths;
    }

    @Override
    public void addOpenflowSliver(@Nonnull OpenflowSliver sliver) {
        //stores BasicOpenflowSliver as we do nothing special with them
        openflowSlivers.add(sliver);
    }

    @Override
    @Nonnull
    public List<OpenflowSliver> getOpenflowSlivers() {
        //stores BasicOpenflowSliver as we do nothing special with them
        return openflowSlivers;
    }

    @Override
    public void addLease(@Nonnull Lease lease) {
        leases.add(lease);
    }

    @Override
    @Nonnull
    public List<Lease> getLeases() {
        return leases;
    }

    @Override
    public void addChannel(@Nonnull Channel channel) {
        assert channel instanceof FXRspecChannel;
        channels.add((FXRspecChannel) channel);
    }

    public boolean deleteChannel(@Nonnull Channel channel) {
        assert channel instanceof FXRspecChannel;
        return channels.remove(channel);
    }

    @Override
    @Nonnull
    public ObservableList<FXRspecChannel> getChannels() {
        return channels;
    }

    @Override
    public void addGeantTestbedType(@Nonnull GeantTestbedType gtt) {
        assert gtt instanceof FXGeantTestbedType;
        geantTestbedTypes.add((FXGeantTestbedType) gtt);
    }

    public boolean deleteGeantTestbedType(@Nonnull GeantTestbedType gtt) {
        assert gtt instanceof FXGeantTestbedType;
        return geantTestbedTypes.remove(gtt);
    }

    @Override
    @Nonnull
    public ObservableList<FXGeantTestbedType> getGeantTestbedTypes() {
        return geantTestbedTypes;
    }

    @Override
    public void addAddressPool(@Nonnull AddressPool addressPool) {
        if (addressPool instanceof FXAddressPool)
            addressPools.add((FXAddressPool) addressPool);
        else
            addressPools.add(new FXAddressPool(addressPool));
    }

    @Override
    public boolean deleteAddressPool(@Nonnull AddressPool addressPool) {
        //noinspection SuspiciousMethodCalls
        return addressPools.remove(addressPool);
    }

    @Override
    @Nonnull
    public ObservableList<FXAddressPool> getAddressPools() {
        return addressPools;
    }

    @Override
    @Nullable
    public FXAddressPool getAddressPoolByClientId(@Nonnull String clientId) {
        return addressPools.stream()
                .filter(addressPool -> Objects.equals(clientId, addressPool.getClientId()))
                .findFirst().orElse(null);

    }

    @Override
    public void addExecuteAnsiblePlaybook(@Nonnull ExecuteAnsiblePlaybook executeAnsiblePlaybook) {
        if (executeAnsiblePlaybook instanceof BasicExecuteAnsiblePlaybook)
            ansibleCookbooks.add((BasicExecuteAnsiblePlaybook) executeAnsiblePlaybook);
        else
            ansibleCookbooks.add(new BasicExecuteAnsiblePlaybook(executeAnsiblePlaybook));
    }

    @Override
    public boolean deleteExecuteAnsiblePlaybook(@Nonnull ExecuteAnsiblePlaybook executeAnsiblePlaybook) {
        //noinspection SuspiciousMethodCalls
        return ansibleCookbooks.remove(executeAnsiblePlaybook);
    }

    @Override
    @Nonnull
    public List<? extends BasicExecuteAnsiblePlaybook> getExecuteAnsiblePlaybooks() {
        return ansibleCookbooks;
    }

    @Override
    public void addDistributeSshKeypair(@Nonnull DistributeSshKeypair distributeSshKeypair) {
        if (distributeSshKeypair instanceof BasicDistributeSshKeypair)
            distributeSshKeypairs.add((BasicDistributeSshKeypair) distributeSshKeypair);
        else
            distributeSshKeypairs.add(new BasicDistributeSshKeypair(distributeSshKeypair));
    }

    @Override
    public boolean deleteDistributeSshKeypair(@Nonnull DistributeSshKeypair distributeSshKeypair) {
        //noinspection SuspiciousMethodCalls
        return distributeSshKeypairs.remove(distributeSshKeypair);
    }

    @Nonnull
    @Override
    public List<DiskImageList> getDiskImageLists() {
        return diskImageLists;
    }

    @Nonnull
    @Override
    public List<FlavorList> getFlavorLists() {
        return flavorLists;
    }

    @Override
    public void addDiskImageList(@Nonnull DiskImageList list) {
        diskImageLists.add(list);
    }

    @Override
    public void addFlavorList(@Nonnull FlavorList list) {
        flavorLists.add(list);
    }

    @Override
    @Nonnull
    public List<? extends DistributeSshKeypair> getDistributeSshKeypairs() {
        return distributeSshKeypairs;
    }

    @Nonnull
    public ObjectProperty<HardwareTypeInfo> hardwareTypeInfoProperty() {
        return hardwareTypeInfo;
    }

    @Nullable
    @Override
    public HardwareTypeInfo getHardwareTypeInfo() {
        return hardwareTypeInfo.get();
    }

    @Override
    public void setHardwareTypeInfo(@Nullable HardwareTypeInfo hardwareTypeInfo) {
        this.hardwareTypeInfo.setValue(hardwareTypeInfo);
    }


    @Override
    @Nullable
    public FXRspecNode getNodeByUniqueId(@Nonnull String id) {
        for (FXRspecNode node : nodes)
            if (Objects.equals(node.getUniqueId(), id))
                return node;
        return null;
    }

    @Override
    @Nullable
    public FXRspecNode getNodeByClientId(@Nonnull String id) {
        for (FXRspecNode node : nodes)
            if (Objects.equals(id, node.getClientId()))
                return node;
        return null;
    }

    @Override
    @Nullable
    public FXRspecNode getNodeByComponentId(@Nonnull String id) {
        for (FXRspecNode node : nodes)
            if (node.getComponentId() != null && Objects.equals(node.getComponentId().toString(), id))
                return node;
        return null;
    }

    @Override
    @Nullable
    public FXRspecChannel getChannelByComponentId(@Nonnull GeniUrn urn) {
        return channels.stream().filter(channel -> Objects.equals(channel.getComponentId(), urn)).findAny().orElse(null);
    }


    @Override
    @Nonnull
    public List<FXRspecNode> getNodesByAuthority(@Nonnull Server authority) {
        assert authority != null;
        List<FXRspecNode> nodesOfAuthority = new ArrayList<>();
        for (FXRspecNode node : nodes) {
            assert node != null;
            if (authority.getDefaultComponentManagerAsGeniUrn().equalsIgnoringSubAuth(node.getComponentManagerId())) {
                nodesOfAuthority.add(node);
            }
        }
        return nodesOfAuthority;
    }

    @Nullable
    public ExperimentBarrierSegment getBarrierSegmentById(int id) {
        for (ExperimentBarrierSegment b : barrierSegments) {
            if (b.getOrderNumber() == id)
                return b;
        }
        return null;
    }

    @Override
    @Nullable
    public FXRspecLink getLinkByClientId(@Nonnull String id) {
        for (FXRspecLink link : links)
            if (Objects.equals(link.getClientId(), id))
                return link;
        return null;
    }

    @Override
    @Nullable
    public FXRspecLink getLinkByUniqueId(@Nonnull String id) {
        for (FXRspecLink link : links)
            if (Objects.equals(link.getUniqueId(), id))
                return link;
        return null;
    }

    @Override
    public void deleteNode(@Nonnull RspecNode node) {
        assert node instanceof FXRspecNode;
        List<RspecLink> linksToDelete = new ArrayList<>();
        List<RspecInterface> ifacesToDelete = new ArrayList<>();
        for (RspecInterface iface : node.getInterfaces()) {
            if (iface.isLinkBound()) {
                RspecLink link = iface.getLink();
                if (link != null && link.getInterfaces().size() <= 2)
                    linksToDelete.add(link);
            }
            ifacesToDelete.add(iface);
        }
        ifacesToDelete.forEach(be.iminds.ilabt.jfed.rspec.model.RspecInterface::delete);
        nodes.remove(node);
        linksToDelete.forEach(this::deleteLink);

        for (ExperimentBarrierSegment b : barrierSegments) {
            b.getCommands().stream()
                    .filter(c -> Objects.equals(c.getNodeUniqueID(), node.getClientId()))
                    .forEach(b::deleteCommand);
        }
    }

    @Override
    public void deleteLink(@Nonnull RspecLink link) {
        assert link instanceof FXRspecLink;
        List<RspecInterface> ifacesToDelete = new ArrayList<>(link.getInterfaces());
        ifacesToDelete.forEach(be.iminds.ilabt.jfed.rspec.model.RspecInterface::delete);
        links.remove(link);
    }

    @Override
    @Nonnull
    public String toGeni3Rspec(@Nonnull RequestRspecSpecialCases specialCase) {
        StaxRspecWriter staxRspecWriter = new StaxRspecWriter(this);

        switch (specialCase) {
            case NONE: {
                //default staxRspecWriter options are fine
                break;
            }
            case PLE: {
                //already using quick hack to get ple to work in: RspecNode.writePropertiesToGeni3RequestRspec()
                List<String> allowedComponentManagers = new ArrayList<>();
                allowedComponentManagers.add("urn:publicid:IDN+ple:ibbtple+authority+cm");
                //hack for ple: don't add nodes from other componentmanagers
                staxRspecWriter.setComponentManagerFilter(allowedComponentManagers);

                //another hack for ple: ple also doesn't like schemaLocation
                staxRspecWriter.setAddSchemaLocation(false);

                //yet another hack for PLE: add a meaningless, empty <sliver/> element to each node
                staxRspecWriter.setAddEmptySliverToNodeHack(true);
                break;
            }
            case NITOS_NO_E_NODE_B_HACK: {
                staxRspecWriter.setUseNitosENodeBHack(false);
                break;
            }
            default:
                throw new RuntimeException("Special case " + specialCase + " not supported");
        }

        try {
            return staxRspecWriter.call();
        } catch (RspecParseException e) {
            LOG.error("Error converting to Rspec: ", e);
            throw new RuntimeException("Error converting to Rspec: " + e, e);
        }
    }

    @Override
    @Nonnull
    public String toGeni3Rspec() {
        return toGeni3Rspec(RequestRspecSpecialCases.NONE);
    }

    @Override
    @Nullable
    public Date getExpires() throws ParseException {
        if (expiresDate.get() == null) return null;
        return RFC3339Util.iso8601StringToDate(expiresDate.get());
    }

    @Override
    public void setExpires(@Nullable Date expires) {
        if (expires == null)
            this.expiresDate.setValue(null);
        else
            this.expiresDate.setValue(RFC3339Util.dateToRFC3339String(expires, true/*forceZulu*/)); //must be Z for emulab
    }

    @Override
    @Nullable
    public String getExpiresString() {
        return expiresDate.get();
    }

    @Override
    public void setExpires(@Nullable String expires) {
        this.expiresDate.setValue(expires);
    }

    @Override
    @Nullable
    public Integer getRoutableAddressesAvailable() {
        return routableAddressesAvailableUsed.getValue() ? routableAddressesAvailable.getValue() : null;
    }

    @Override
    @Nullable
    public Integer getRoutableAddressesConfigured() {
        return routableAddressesConfiguredUsed.getValue() ? routableAddressesConfigured.getValue() : null;
    }

    @Override
    public void setRoutableAddressesAvailable(@Nullable Integer i) {
        if (i == null) {
            routableAddressesAvailableUsed.set(false);
        } else {
            routableAddressesAvailableUsed.set(true);
            routableAddressesAvailable.set(i);
        }
    }

    @Override
    public void setRoutableAddressesConfigured(@Nullable Integer i) {
        if (i == null) {
            routableAddressesConfiguredUsed.set(false);
        } else {
            routableAddressesConfiguredUsed.set(true);
            routableAddressesConfigured.set(i);
        }
    }

    @Override
    @Nullable
    public FXGeantTestbedType getGeantTestbedTypeByName(@Nonnull String geantTestbedTypeName) {
        return geantTestbedTypes.stream()
                .filter(gtt -> Objects.equals(gtt.getName(), geantTestbedTypeName))
                .findFirst().orElse(null);
    }

    /**
     * This gathers all the component_manager_id's mentioned in the RSpec.
     *
     * If a new type is added here, it should also be added to the HasComponentManagerIdsValidator to prevent
     * invalid Rspecs
     */
    @Override
    @Nonnull
    public List<GeniUrn> getAllComponentManagerUrns() {
        Set<GeniUrn> urns = new HashSet<>();

        nodes.stream().map(RspecNode::getComponentManagerId).filter(Objects::nonNull).forEach(urns::add);
        links.stream().flatMap(link -> link.getComponentManagerUrns().stream()).forEach(urns::add);
        channels.stream().map(Channel::getComponentManagerId).forEach(urns::add);
        addressPools.stream().map(AddressPool::getComponentManagerId).forEach(urns::add);
        geantTestbedTypes.stream().map(GeantTestbedType::getComponentManagerId).forEach(urns::add);

        //make sure that a null value is removed! It shouldn't be there in the first place, but better safe than sorry.
        if (urns.contains(null)) {
            LOG.warn("Detected a null component_manager_id. This rspec is not valid! Doing a workaround now...");
            urns.remove(null);
        }

        return new ArrayList<>(urns);
    }

    @Override
    @Nonnull
    public ObservableList<UserSpec> getSshKeys() {
        return sshKeys;
    }

    @Override
    public void setSshKeys(@Nonnull List<UserSpec> newKeys) {
        sshKeys.setAll(newKeys);
    }

    @Override
    @Nonnull
    public String nextLinkName() {
        int c = links.size();
        String res = "link" + c;
        while (getLinkByClientId(res) != null)
            res = "link" + (++c);

        return res;
    }

    @Override
    @Nonnull
    public String nextNodeName() {
        int c = nodes.size();
        String res = "node" + c;
        while (getNodeByClientId(res) != null)
            res = "node" + (++c);

        return res;
    }

    @Override
    @Nullable
    public String getType() {
        return rspecType;
    }

    @Override
    @Nonnull
    public String nextIfaceName(@Nonnull RspecNode node) {
        int c = node.getInterfaces().size();
        String res = node.getClientId() + ":if" + c;
        while (getInterfaceByClientId(res) != null)
            res = node.getClientId() + ":if" + (++c);

        return res;
    }

    @Override
    @Nullable
    public FXRspecInterface getInterfaceByUniqueId(@Nonnull String uniqueId) {
        for (FXRspecNode node : nodes) {
            FXRspecInterface iface = node.getInterfaceByUniqueId(uniqueId);
            if (iface != null)
                return iface;
        }
        for (FXRspecLink link : links) {
            FXRspecInterface iface = link.getInterfaceByUniqueId(uniqueId);
            if (iface != null)
                return iface;
        }
        return null;
    }

    @Override
    @Nullable
    public FXRspecInterface getInterfaceByClientId(@Nonnull String clientId) {
        FXRspecInterface res = nodes.stream()
                .map(node -> node.getInterfaceByClientId(clientId))
                .filter(Objects::nonNull)
                .findFirst().orElse(null);
        if (res == null) {
            res = links.stream()
                    .map(link -> link.getInterfaceByClientId(clientId))
                    .filter(Objects::nonNull)
                    .findFirst().orElse(null);
        }
        return res;
    }

    @Override
    @Nullable
    public Integer getMultiplexFactor() {
        return multiplexFactor.get();
    }

    @Nonnull
    public ObjectProperty<Integer> multiplexFactorProperty() {
        return multiplexFactor;
    }

    public void setMultiplexFactor(@Nullable Integer multiplexFactor) {
        this.multiplexFactor.set(multiplexFactor);
    }

    @Override
    @Nonnull
    public List<ExtraXml> getExtraXml() {
        return extraXml;
    }

    @Override
    public void setExtraXml(@Nonnull List<ExtraXml> extraXml) {
        this.extraXml.clear();
        this.extraXml.addAll(extraXml);
    }

    @Override
    @Nonnull
    public ModelRspec copy() {
        return factory.copyModelRspec(this, null/*new rspec type (request/manifest/...)*/);
    }
}
